# three-pathfinding

Navigation mesh toolkit for ThreeJS, based on [PatrolJS](https://github.com/nickjanssen/PatrolJS). Computes paths between points on a 3D nav mesh, supports multiple zones, and clamps movement vectors for FPS controls. To learn how to create a navigation mesh using Blender, see [Creating a Nav Mesh](https://www.donmccurdy.com/2017/08/20/creating-a-nav-mesh-for-a-webvr-scene/).

Thanks to [Nick Janssen](https://github.com/nickjanssen) for creating [PatrolJS](https://github.com/nickjanssen/PatrolJS), which was the basis for this library.

![screenshot](https://user-images.githubusercontent.com/1848368/34424850-d79e5a24-ebf4-11e7-87c4-afc75cdc41bd.png)

## Introduction

Traditionally games and 3D apps used waypoints to help their AI agents navigate. This is bad and has a lot of problems, but is generally easier to implement than navigation meshes. Navmeshes are far more accurate, faster, and take into account the size of the AI agent (e.g. tanks require move space to maneuver than soldiers).

For a thorough introduction to Navigation mesh pathfinding, see AI Blog's article, [Fixing Pathfinding Once and For All](http://www.ai-blog.net/archives/000152.html).

## Quickstart

### Installation

```
npm install --save three-pathfinding
```

### Creating a Navigation Mesh

This library does not build navigation meshes for you — instead, create a navigation mesh using [Blender](https://youtu.be/v4d_6ZCGlAg?t=6m8s), [Recast](https://github.com/recastnavigation/recastnavigation) ([CLI](https://github.com/but0n/recastCLI.js)), or another tool.

The library accepts a [THREE.Geometry](https://threejs.org/docs/#api/core/Geometry) instance, which can be loaded with any three.js loader and converted from BufferGeometry if necessary.

### Example

Loading a mesh from a `.gltf` file:

```js
let mesh;

const loader = new THREE.GLTFLoader();
loader.load( 'navmesh.gltf', ({scene}) => {
    scene.traverse((node) => {
        if (node.isMesh) mesh = node;
    });
}, undefined, (e) => {
    console.error(e);
});
```

Initializing the library, creating a level, and finding a path:

```js
// ES6
import { Pathfinding } from 'three-pathfinding';
// CommonJS
const Pathfinding = require('three-pathfinding').Pathfinding;
// UMD
const Pathfinding = window.threePathfinding.Pathfinding;

// Create level.
const pathfinding = new Pathfinding();
const ZONE = 'level1';
pathfinding.setZoneData(ZONE, Pathfinding.createZone(mesh.geometry));

// Find path from A to B.
const groupID = pathfinding.getGroup(ZONE, a);
const path = pathfinding.findPath(a, b, ZONE, groupID);
```

### Running the demo locally

```
git clone https://github.com/donmccurdy/three-pathfinding.git
cd three-pathfinding

npm install
npm run dev
```

The demo will start at http://localhost:9966/demo/demo.html.

## API

<!--- API BEGIN --->

<!-- Generated by documentation.js. Update this documentation by updating the source code. -->

### Table of Contents

-   [Pathfinding][1]
    -   [setZoneData][2]
    -   [getGroup][3]
    -   [getRandomNode][4]
    -   [getClosestNode][5]
    -   [findPath][6]
    -   [clampStep][7]
    -   [createZone][8]
-   [Zone][9]
-   [Group][10]
-   [Node][11]

## Pathfinding

Defines an instance of the pathfinding module, with one or more zones.

### setZoneData

Sets data for the given zone.

**Parameters**

-   `zoneID` **[string][12]** 
-   `zone` **[Zone][13]** 

### getGroup

Returns closest node group ID for given position.

**Parameters**

-   `zoneID` **[string][12]** 
-   `position` **THREE.Vector3** 

Returns **[number][14]** 

### getRandomNode

Returns a random node within a given range of a given position.

**Parameters**

-   `zoneID` **[string][12]** 
-   `groupID` **[number][14]** 
-   `nearPosition` **THREE.Vector3** 
-   `nearRange` **[number][14]** 

Returns **[Node][15]** 

### getClosestNode

Returns the closest node to the target position.

**Parameters**

-   `position` **THREE.Vector3** 
-   `zoneID` **[string][12]** 
-   `groupID` **[number][14]** 
-   `checkPolygon` **[boolean][16]**  (optional, default `false`)

Returns **[Node][15]** 

### findPath

Returns a path between given start and end points. If a complete path
cannot be found, will return the nearest endpoint available.

**Parameters**

-   `startPosition` **THREE.Vector3** Start position.
-   `targetPosition` **THREE.Vector3** Destination.
-   `zoneID` **[string][12]** ID of current zone.
-   `groupID` **[number][14]** Current group ID.

Returns **[Array][17]&lt;THREE.Vector3>** Array of points defining the path.

### clampStep

Clamps a step along the navmesh, given start and desired endpoint. May be
used to constrain first-person / WASD controls.

**Parameters**

-   `start` **THREE.Vector3** 
-   `end` **THREE.Vector3** Desired endpoint.
-   `node` **[Node][15]** 
-   `zoneID` **[string][12]** 
-   `groupID` **[number][14]** 
-   `endTarget` **THREE.Vector3** Updated endpoint.

Returns **[Node][15]** Updated node.

### createZone

(Static) Builds a zone/node set from navigation mesh geometry.

**Parameters**

-   `geometry` **THREE.Geometry** 

Returns **[Zone][13]** 

## Zone

Defines a zone of interconnected groups on a navigation mesh.

**Properties**

-   `groups` **[Array][17]&lt;[Group][18]>** 

## Group

Defines a group within a navigation mesh.

## Node

Defines a node (or polygon) within a group.

**Properties**

-   `id` **[number][14]** 
-   `neighbours` **[Array][17]&lt;[number][14]>** IDs of neighboring nodes.
-   `centroid` **THREE.Vector3** 
-   `portals` **[Array][17]&lt;[Array][17]&lt;[number][14]>>** Array of portals, each defined by two vertex IDs.
-   `closed` **[boolean][16]** 
-   `cost` **[number][14]** 

[1]: #pathfinding

[2]: #setzonedata

[3]: #getgroup

[4]: #getrandomnode

[5]: #getclosestnode

[6]: #findpath

[7]: #clampstep

[8]: #createzone

[9]: #zone

[10]: #group

[11]: #node

[12]: https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/String

[13]: #zone

[14]: https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Number

[15]: #node

[16]: https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Boolean

[17]: https://developer.mozilla.org/docs/Web/JavaScript/Reference/Global_Objects/Array

[18]: #group
<!--- API END --->

## Thanks to

* [PatrolJS](https://github.com/nickjanssen/PatrolJS)
* [bgrin's astar library](https://github.com/bgrins/javascript-astar)
* [Digesting Duck's Simple Stupid Funnel Algorithm](http://digestingduck.blogspot.jp/2010/03/simple-stupid-funnel-algorithm.html)
* [Recastnavigation's level mesh](https://github.com/memononen/recastnavigation)
* [Constrained Movement Along Navmesh pt. 3](http://digestingduck.blogspot.com/2010/07/constrained-movement-along-navmesh-pt-3.html?m=1)
